local dmg = module.internal("damage")
local ts = module.internal("TS")
local pred = module.internal("pred")
local orb = module.internal("orb")

local spell_parent = module.load("GG-AIO", "spell")
local spell = class(spell_parent)

local common = module.load("GG-AIO","common")

local getTar = ts.get_result
spell.slot = SpellSlot.Q
spell.pred = {
    range = player.attackRange + player.boundingRadius,
    speed = 1400,
    angle = 160*math.pi/180,
    delay = 0.25,
}
spell.cone_length = 500
spell.wide_cone_length = 150
spell.target = nil
spell.time = 0
spell.kalman = module.load("GG-AIO", "kalman")()
spell.kalman_ms_high = 750
spell.kalman_ms_mid = 500
spell.kalman_ms_low = 325



function spell:__init(menu)
    self:__super(menu)
end
-- add draw q killable
--q clear logic : kill 2 minions
function spell:makeMenu()
    self.menu:set("texture",player:spellSlot(self.slot).sprite)
    self.menu:header("qcombo","Combo")
    self.menu:boolean("mfuseqcombo","Use Q in Combo", true)
    self.menu:header("qharass","Harass")
    self.menu:boolean("mfuseqharass", "Use Q in Harass", true)
    self.menu:boolean("mfautoqharass", "Auto harass", false)
    self.menu:boolean("mfqkillharass", "Only use Q if minion will die", false)
    self.menu:slider("mfqharassmana","Minimum mana percentage needed", 30, 0, 100, 1)
    self.menu:header("qclear","Lane Clear")
    self.menu:boolean("mfuseqclear","Use Q in lane clear", true)
    self.menu:slider("mfqclearmana", "Minimum mana percentage needed", 30, 0, 100, 1)
    self.menu:header("qmisc", "Misc")
    self.menu:boolean("mfqdrawkillable", "Draw killable with q", false)
end

function spell:qPredFilter(target)
    -- WIP, needs fine tuning
    if pred.trace.newpath(target, 0.022, 0.185) and self.kalman:hasVelocity(self.kalman_ms_high, target, 1350) then
        return true
    elseif pred.trace.newpath(target, 0.022, 0.310) and self.kalman:hasVelocity(self.kalman_ms_high, target, 550) then
        return true
    elseif pred.trace.newpath(target, 0.015, 0.095) and self.kalman:hasVelocity(self.kalman_ms_high, target, 50) then
        return true
    elseif pred.trace.newpath(target, 0.033, 0.245) and self.kalman:hasVelocity(self.kalman_ms_mid, target, 1350) then
        return true
    elseif pred.trace.newpath(target, 0.033, 0.370) and self.kalman:hasVelocity(self.kalman_ms_mid, target, 550) then
        return true
    elseif pred.trace.newpath(target, 0.010, 0.105) and self.kalman:hasVelocity(self.kalman_ms_mid, target, 50) then
        return true
    elseif pred.trace.newpath(target, 0.033, 0.305) and self.kalman:hasVelocity(self.kalman_ms_low, target, 1350) then
        return true
    elseif pred.trace.newpath(target, 0.033, 0.550) and self.kalman:hasVelocity(self.kalman_ms_low, target, 550) then
        return true
    elseif pred.trace.newpath(target, 0.010, 0.125) and self.kalman:hasVelocity(self.kalman_ms_low, target, 50) then
        return true
    elseif pred.trace.newpath(target, 0.025, 0.200) and self.kalman:hasVelocity(self.kalman_ms_low, target, 200) then
        return true
    elseif self.kalman.velocity[target.networkID] < 50 and 1000 * game.time - self.kalman.lastboost[target.networkID] > 500 then
        return true
    elseif self.kalman.velocity[target.networkID] < self.kalman_ms_high and 1000 * game.time - self.kalman.lastboost[target.networkID] < 55 then
        return true
    elseif self.kalman.velocity[target.networkID] < self.kalman_ms_mid and 1000 * game.time - self.kalman.lastboost[target.networkID] < 75 then
        return true
    end
end

local function normalFilter(res, tar, dist)
    -- returns the best target in range
    if dist > spell.pred.range then
        return false
    end
    res.obj = tar
    return true
end

function spell:updateTarget()
    -- updates self.target with the best target
    local res = getTar(normalFilter, nil, false, true)
    if not res or not res.obj then
        return
    end
    if res.obj:isValidTarget() then 
        self.target = res.obj
    end
end

function spell:updateRange()
    self.pred.range = player.attackRange + player.boundingRadius
end

function spell:bounce(target,obj_table) -- obj_table : valid target for q : minions and hero
    -- returns which gameobject will be hit by the q bounce on target
    local pt_vec = target.pos - player.pos
    local deg20table = {}
    local deg40table = {}
    local deg110table = {}
    local deg160table = {} -- within 150 units in a 160 deg cone
    for _,obj in pairs(obj_table) do
        if obj ~= target then
            local dist = obj:dist(target.pos)
            local delay = self.pred.delay + target:dist()/self.pred.speed
            local future_pos = pred.core.get_pos_after_time(obj,delay)
            local dist_future = future_pos:dist(target.pos)
            if dist < self.cone_length and dist_future < self.cone_length then
                local to_vec = future_pos - target.pos
                local minion_angle = math.abs(pt_vec:angle(to_vec))
                if minion_angle < 10 then
                    table.insert(deg20table, obj)
                elseif minion_angle < 20 then
                    table.insert(deg40table, obj)
                elseif minion_angle < 55 then
                    table.insert(deg110table,obj)
                --elseif dist < self.wide_cone_length and minion_angle < 80 then
                --    table.insert(deg160table, obj)
                end
            end
        end
    end
    if next(deg20table) ~= nil then
        local closest = common:getClosestObj(target,deg20table)
        return closest
    elseif next(deg40table) ~= nil then
        local closest = common:getClosestObj(target,deg40table)
        return closest
    elseif next(deg110table) ~= nil then
        local closest = common:getClosestObj(target,deg110table)
        return closest
    --elseif next(deg160table) ~= nil then
    --    local closest = common:getClosestObj(target,deg160table)
    --    return closest
    else
        return nil --nobody in the cone :(
    end
end
-- improve synergy with orb
function spell:passiveDamage(target)
    local damage
    if player.levelRef < 4 then
        damage = player.totalBonusAttackDamage * 0.5
    elseif player.levelRef < 7 then
        damage = player.totalBonusAttackDamage * 0.6
    elseif player.levelRef < 9 then
        damage = player.totalBonusAttackDamage * 0.7
    elseif player.levelRef < 11 then
        damage = player.totalBonusAttackDamage * 0.8
    elseif player.levelRef < 13 then
        damage = player.totalBonusAttackDamage * 0.9
    else
        damage = player.totalBonusAttackDamage
    end
    return damage
end

function spell:bouncehit(obj_table)
    if self:usable() then
        for _,obj in pairs(obj_table) do
            if obj.isMinion and obj:dist()<self.pred.range then
                local dist = obj:dist()
                local delay = self.pred.delay + obj:dist()/self.pred.speed
                local health_pred = dmg.predict(obj,delay)
                local dmg_spell = dmg.spell(player,obj, self.slot)
                local cond1 = dmg_spell > health_pred 
                local cond2 = (dmg_spell + dmg.autoattack(player, obj, true) - self:passiveDamage(obj) > health_pred) and dmg.autoattack(player, obj, true) < health_pred
                local double = self:bounce(obj,obj_table)
                if double and double.isHero then
                    print(self:qPredFilter(double))
                    if cond2 and orb.core.can_attack() then
                        player:attack(obj)
                    elseif orb.core.can_spell(0.2) and self:qPredFilter(double) then
                        if self.menu.mfqkillharass:get() then
                            if cond1 then
                                player:castSpell("obj",self.slot,obj)
                            end
                        else
                            player:castSpell("obj",self.slot,obj)
                        end
                    end
                end
            end
        end
    end
end

function spell:castAfterAA()
    if orb.core.get_target() and orb.core.get_target().isHero and orb.core.can_spell_after_attack(0,true) and self:usable() then
        player:castSpell("obj",self.slot,orb.core.get_target())
    end
end

function spell:harass(obj_table)
    if not self.menu.mfautoqharass:get() then
        self:bouncehit(obj_table)
    end
    self:castAfterAA()
end

function spell:combo()
    self:castAfterAA()
end

function spell:clear(obj_table)
    if self:usable() then
        for _,obj in pairs(obj_table) do
            if obj:dist() < self.pred.range then
                local double = self:bounce(obj,obj_table)
                if double and double.isMinion then
                    if (dmg.spell(player,double,self.slot) > double.health) or (dmg.spell(player,obj, self.slot) > obj.health and dmg.spell(player,double,self.slot)*2 > double.health) then -- it doesn't take into account ie
                        player:castSpell("obj", self.slot, obj)
                    end
                end
            end
        end
    end
end

function spell:draw(obj_table)
    if self.menu.mfqdrawkillable:get() then
        for _,obj in pairs(obj_table) do
            if obj.isMinion and obj:isValidTarget() and (dmg.spell(player,obj,self.slot) > obj.health) then
                graphics.draw_circle_filled(obj.pos, 50, 0xffc90000)
            end
        end
    end
end

return spell
