-- Thanks dieno !

local kalman_filter = class()

function kalman_filter:__init()
    self.current_state_estimate = 0
    self.current_prob_estimate = 0
    self.Q = 1
    self.R = 15
end
function kalman_filter:step(control_vector, measurement_vector)
    local predicted_state_estimate = self.current_state_estimate + control_vector
    local predicted_prob_estimate = self.current_prob_estimate + self.Q
    local innovation = measurement_vector - predicted_state_estimate
    local innovation_covariance = predicted_prob_estimate + self.R
    local kalman_gain = predicted_prob_estimate / innovation_covariance
    self.current_state_estimate = predicted_state_estimate + kalman_gain * innovation
    self.current_prob_estimate = (1 - kalman_gain) * predicted_prob_estimate
    return self.current_state_estimate
end

local kalman = class()

kalman.eneplayeres = {}
kalman.kalmanFilters = {}
kalman.velocityTimers = {}
kalman.oldPosx = {}
kalman.oldPosz = {}
kalman.oldTick = {}
kalman.velocity = {}
kalman.lastboost = {}
kalman.velocity_TO = 10
kalman.CONVERSATION_FACTOR = 975
kalman.MS_MIN = 325 -- default value, may need to be changed, used in "update velocity"

function kalman:__init()
    for hero in objManager.ienemies do
        table.insert(self.eneplayeres, hero)
        self.kalmanFilters[hero.networkID] = kalman_filter()
        self.velocityTimers[hero.networkID] = 0
        self.oldPosx[hero.networkID] = 0
        self.oldPosz[hero.networkID] = 0
        self.oldTick[hero.networkID] = 0
        self.velocity[hero.networkID] = 0
        self.lastboost[hero.networkID] = 0
    end
end

function kalman:hasVelocity(velocity,target,time)
    if target and target:isValidTarget() then
        return (self.velocity[target.networkID] < velocity and target.moveSpeed < velocity and 1000 * game.time - self.lastboost[target.networkID] > time)
    else
        return nil
    end
end

function kalman:_calcHeroVelocity(target, oldPosx, oldPosz, oldTick)
    if oldPosx and oldPosz and target.x and target.z then
        local dis = math.sqrt((oldPosx - target.pos.x) ^ 2 + (oldPosz - target.pos.z) ^ 2)
        self.velocity[target.networkID] = self.kalmanFilters[target.networkID]:step(0, (dis / (1000 * game.time - oldTick)) * self.CONVERSATION_FACTOR)
    end
end

function kalman:UpdateSpeed()
    local tick = 1000 * game.time
    for hero in objManager.ienemies do
        if self.velocityTimers[hero.networkID] <= tick and hero and hero.pos.x and hero.pos.z and (tick - self.oldTick[hero.networkID]) > (self.velocity_TO - 1) then
            self.velocityTimers[hero.networkID] = tick + self.velocity_TO
            self:_calcHeroVelocity(hero, self.oldPosx[hero.networkID], self.oldPosz[hero.networkID], self.oldTick[hero.networkID])
            self.oldPosx[hero.networkID] = hero.x
            self.oldPosz[hero.networkID] = hero.z
            self.oldTick[hero.networkID] = tick
            if self.velocity[hero.networkID] > self.MS_MIN then
                self.lastboost[hero.networkID] = tick
            end
        end
    end
end

return kalman